using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Net;
using System.Net.Sockets;
using AITSW.VersatileSerializing.Customizing;
using System.Diagnostics;

/*
 * References:
 * X.690 (BER): http://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
 * Standards Overview: http://asn1.elibel.tm.fr/en/standards/
 * X.690 Amendmend 1: http://www.itu.int/ITU-T/studygroups/com17/languages/X690amd1.pdf
 * RFC1155: http://www.apps.ietf.org/rfc/rfc1155.html
 * RFC1905: http://rfc.net/rfc1905.html
 * Wikipedia Article: http://en.wikipedia.org/wiki/Simple_Network_Management_Protocol
 * German Wikipedia Article: http://de.wikipedia.org/wiki/Simple_Network_Management_Protocol
 */

namespace AITSW.VersatileSerializing.SNMP
{
#pragma warning disable 414 //warning field is assigned but its value is never used

    /// <summary>
    /// Defines the tags used to identify a BER type.
    /// </summary>
    /// <remarks>
    /// Note that this is not a comprehensive list. It just represents the most commonly encountered
    /// tags. Examples of currently unsupported tags, that might be added later, are Opaque (0x44),
    /// Counter64 (0x46), UInteger32 (0x47), Integer64 (0x4A), UInteger64 (0x4B) and probably others.
    /// </remarks>
    public enum BER_Tag : byte
    {
        Boolean = 1,
        Integer = 2,
        OctetString = 4,
        Null = 5,
        ObjId = 6,
        Sequence = 0x30,
        IpAddress = 0x40,
        Counter32 = 0x41,
        Gauge32 = 0x42,
        TimeTicks = 0x43,
        NoSuchObject = 0x80,
        NoSuchInstance = 0x81,
        EndOfMIBView = 0x82
    }

    public enum SNMP_ErrorStatus : int
    {
        noError = 0,
        tooBig = 1,
        noSuchName = 2,
        badValue = 3,
        readOnly = 4,
        genErr = 5,
        noAccess = 6,
        wrongType = 7,
        wrongLength = 8,
        wrongEncoding = 9,
        wrongValue = 10,
        noCreation = 11,
        inconsistentValue = 12,
        resourceUnavailable = 13,
        commitFailed = 14,
        undoFailed = 15,
        authorizationError = 16,
        notWritable = 17,
        inconsistentName = 18
    }

    public class BER_Length : VSCustomizableNumBase<uint>
    {
        public BER_Length()
            : base()
        {
        }
        public BER_Length(uint initValue)
            : base(initValue)
        {
        }

        public override bool FromBinBytes(VSCustomizationContext context, byte[] data, int startIndex, int? byteLenHint, out int bytesUsed)
        {
            int bytes_avail = data.Length - startIndex;
            uint u1, u2;
            bytesUsed = 0;

            if (bytes_avail <= 0) return false;
            u1 = Convert.ToUInt32(data[startIndex]);
            bytes_avail--;
            bytesUsed++;
            if (u1 < 0x80)
                this.value = u1;
            else
            {
                u2 = 0;
                u1 &= 0x7F;
                if (bytes_avail < u1) return false;
                for (int i = 1; i <= u1; i++)
                {
                    u2 = u2 * 256 + Convert.ToUInt32(data[startIndex + i]);
                    bytesUsed++;
                }
                if ((byteLenHint != null) && (bytesUsed != byteLenHint)) return false;
                this.value = u2;
            }

            return true;
        }
        public override int? GetConstantBinByteSize(VSCustomizationContext context)
        {
            return null;
        }
        public override byte[] ToBytes(VSCustomizationContext context)
        {
            if (Value < 0x80)
                return new byte[1] { Convert.ToByte(Value) };
            else
            {
                List<byte> bl = new List<byte>();
                uint u = Value;
                do
                {
                    bl.Insert(0, Convert.ToByte(u % 256));
                    u /= 256;
                }
                while (u != 0);
                bl.Insert(0, Convert.ToByte(bl.Count | 0x80));
                return bl.ToArray();
            }
        }
    }

    public class BER_Boolean : VSCustomizableTypeBase<bool>
    {
        public override bool FromBinBytes(VSCustomizationContext context, byte[] data, int startIndex, int? byteLenHint, out int bytesUsed)
        {
            bytesUsed = 0;
            if ((byteLenHint == null) || (byteLenHint < 1)) return false;

            bytesUsed = 1;
            this.value = data[startIndex] != 0;
            return true;
        }
        public override int? GetConstantBinByteSize(VSCustomizationContext context)
        {
            return 1;
        }
        public override byte[] ToBytes(VSCustomizationContext context)
        {
            return new byte[] { (this.value ? (byte)1 : (byte)0) };
        }
    }

    public class BER_Integer : VSCustomizableNumBase<int>
    {
        public override bool FromBinBytes(VSCustomizationContext context, byte[] data, int startIndex, int? byteLenHint, out int bytesUsed)
        {
            bytesUsed = 0;
            if ((byteLenHint == null) || (byteLenHint < 1)) return false;

            int val = 0;
            byte b;
            bool neg = false;

            for (int i = 0; i < byteLenHint; i++)
            {
                b = data[startIndex + i];
                if ((i == 0) && (b > 0x7f)) neg = true;
                if (neg) b = (byte)~b;
                val = val * 256 + b;
            }
            if (neg) val = -(val + 1);
            bytesUsed = (int)byteLenHint;
            this.value = val;
            return true;
        }
        public override int? GetConstantBinByteSize(VSCustomizationContext context)
        {
            return null;
        }
        public override byte[] ToBytes(VSCustomizationContext context)
        {
            bool is_neg = Value < 0;
            uint u = (uint)Math.Abs(Value);
            if (is_neg) u = ~(u - 1);
            List<byte> bl = new List<byte>();
            byte b = 0;
            do
            {
                bl.Insert(0, b = Convert.ToByte(u % 256));
                u /= 256;
            }
            while (u != 0);

            if (!is_neg && (b >= 0x80)) bl.Insert(0, 0x00);
            return bl.ToArray();
        }
    }

    public class BER_UInteger : VSCustomizableNumBase<uint>
    {
        public override bool FromBinBytes(VSCustomizationContext context, byte[] data, int startIndex, int? byteLenHint, out int bytesUsed)
        {
            bytesUsed = 0;
            if ((byteLenHint == null) || (byteLenHint < 1)) return false;
            uint val = 0;
            for (int i = 0; i < byteLenHint; i++)
                val = val * 256 + data[startIndex + i];
            bytesUsed = (int)byteLenHint;
            this.value = val;
            return true;
        }
        public override int? GetConstantBinByteSize(VSCustomizationContext context)
        {
            return null;
        }
        public override byte[] ToBytes(VSCustomizationContext context)
        {
            uint u = Value;
            List<byte> bl = new List<byte>();
            do
            {
                bl.Insert(0, Convert.ToByte(u % 256));
                u /= 256;
            }
            while (u != 0);
            return bl.ToArray();
        }
    }

    public class BER_ObjId : VSCustomizableStringBase
    {
        public override bool FromBinBytes(VSCustomizationContext context, byte[] data, int startIndex, int? byteLenHint, out int bytesUsed)
        {
            bytesUsed = 0;
            if ((byteLenHint == null) || (byteLenHint < 1)) return false;

            StringBuilder sb = new StringBuilder();
            int index = startIndex, max_index = startIndex + (int)byteLenHint - 1;

            uint? id;
            id = decodeOIDItem(data, ref index, max_index);
            if (id == null) return false;
            sb.Append(id / 40);
            sb.Append('.');
            sb.Append(id % 40);
            while (index <= max_index)
            {
                id = decodeOIDItem(data, ref index, max_index);
                if (id == null) return false;
                sb.Append('.');
                sb.Append(id);
            }

            this.value = sb.ToString();
            bytesUsed = index - startIndex;
            return true;
        }
        public override byte[] ToBytes(VSCustomizationContext context)
        {
            string[] sl = value.Split('.');
            if (sl.Length < 2) throw new ApplicationException("Unexpected string format for BER_ObjId");
            uint[] ids = new uint[sl.Length];
            for (int i = 0; i < sl.Length; i++)
            {
                if (!uint.TryParse(sl[i], out ids[i]))
                    throw new ApplicationException("Unexpected string format for BER_ObjId");
            }

            List<byte> bl = encodeOIDItem(ids[0] * 40 + ids[1]);
            for (int i = 2; i < ids.Length; i++)
                bl.AddRange(encodeOIDItem(ids[i]));
            return bl.ToArray();
        }
        private uint? decodeOIDItem(byte[] data, ref int index, int maxIndex)
        {
            uint u1 = 0, u2;
            int idx = index;
            while (true)
            {
                u1 *= 128;
                if (idx > maxIndex) return null;
                u2 = data[idx++];
                u1 += u2 & 0x7F;
                if (u2 < 0x80) break;
            }

            index = idx;
            return u1;
        }
        private List<byte> encodeOIDItem(uint val)
        {
            List<byte> bl = new List<byte>();

            bool b = false;
            uint xm;
            do
            {
                xm = val % 128;
                val /= 128;
                if (b) xm |= 0x80;
                if (val > 0) b = true;
                bl.Insert(0, Convert.ToByte(xm));
            }
            while (val != 0);

            return bl;
        }
    }

    public abstract class ASN1_Base
    {
        public virtual bool IsErrorIndicator
        {
            get { return false; }
        }
    }

    [VSSerializableContainer]
    public abstract class ASN1_EmptyBase : ASN1_Base
    {
        [VSSerializableBinary(100)]
        protected BER_Length zero_length = new BER_Length(0);
    }

    [VSSerializableContainer]
    public abstract class ASN1_ValueBase : ASN1_Base
    {
        [VSSerializableBinary(100, CounterMode = VSCounterMode.BytesStored)]
        protected BER_Length length = new BER_Length();
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Boolean)]
    public class ASN1_Boolean : ASN1_ValueBase
    {
        public ASN1_Boolean()
            : base()
        {
        }
        public ASN1_Boolean(bool initValue)
            : base()
        {
            this.BoolVal.Value = initValue;
        }

        [VSSerializableBinary(100)]
        public BER_Boolean BoolVal = new BER_Boolean();

        public override string ToString()
        {
            return this.BoolVal.ToString();
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Integer)]
    public class ASN1_Integer : ASN1_ValueBase
    {
        public ASN1_Integer()
            : base()
        {
        }
        public ASN1_Integer(int initValue)
            : base()
        {
            this.IntVal.Value = initValue;
        }

        [VSSerializableBinary(100)]
        public BER_Integer IntVal = new BER_Integer();

        public override string ToString()
        {
            return this.IntVal.ToString();
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.OctetString)]
    public class ASN1_OctetString : ASN1_ValueBase
    {
        public ASN1_OctetString()
            : base()
        {
        }
        public ASN1_OctetString(string initValue)
            : base()
        {
            this.StrVal = initValue;
        }

        [VSSerializableBinary(100, OmitIfNullOrEmpty = true)]
        private byte[] strData;

        public byte[] StrData
        {
            get { return this.strData; }
            set { this.strData = value; }
        }

        public string StrVal
        {
            get
            {
                if (this.strData == null)
                    return null;
                else
                    return Encoding.GetEncoding("ISO-8859-15").GetString(this.strData);
            }
            set
            {
                if (value == null)
                    this.strData = null;
                else
                    this.strData = Encoding.GetEncoding("ISO-8859-15").GetBytes(value);
            }
        }

        public override string ToString()
        {
            if ((this.strData == null) || (this.strData.Length == 0)) return "(zero-length)";
            else
            {
                bool readable = true;
                for (int i = 0; i < this.strData.Length; i++)
                    if (strData[i] < 32)
                    {
                        readable = false;
                        break;
                    }
                if (readable)
                    return this.StrVal;
                else
                    return BitConverter.ToString(this.strData);
            }
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Null)]
    public class ASN1_Null : ASN1_EmptyBase
    {
        public override string ToString()
        {
            return "NULL";
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.ObjId)]
    public class ASN1_ObjId : ASN1_ValueBase
    {
        public ASN1_ObjId()
            : base()
        {
        }
        public ASN1_ObjId(string oid)
            : base()
        {
            this.OID.Value = oid;
        }

        [VSSerializableBinary(100)]
        public BER_ObjId OID = new BER_ObjId();

        public override string ToString()
        {
            return this.OID.ToString();
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.IpAddress)]
    public class ASN1_IpAddress : ASN1_ValueBase
    {
        public ASN1_IpAddress()
            : base()
        {
        }
        public ASN1_IpAddress(uint initValue)
            : base()
        {
            this.ipAddress.Value = initValue;
        }

        [VSSerializableBinary(100)]
        private BER_UInteger ipAddress = new BER_UInteger();

        public IPAddress IPAddress
        {
            get
            {
                if (this.ipAddress != null)
                {
                    byte[] ba = BitConverter.GetBytes(this.ipAddress.Value);
                    Array.Reverse(ba);
                    return new IPAddress(ba);
                }
                else
                    return IPAddress.None;
            }
            set
            {
                byte[] ba = value.GetAddressBytes();
                Array.Reverse(ba);
                this.ipAddress.Value = BitConverter.ToUInt32(ba, 0);
            }
        }

        public override string ToString()
        {
            return this.IPAddress.ToString();
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Counter32)]
    public class ASN1_Counter32 : ASN1_ValueBase
    {
        public ASN1_Counter32()
            : base()
        {
        }
        public ASN1_Counter32(uint initValue)
            : base()
        {
            this.Counter.Value = initValue;
        }

        [VSSerializableBinary(100)]
        public BER_UInteger Counter = new BER_UInteger();

        public override string ToString()
        {
            return this.Counter.ToString();
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Gauge32)]
    public class ASN1_Gauge32 : ASN1_ValueBase
    {
        public ASN1_Gauge32()
            : base()
        {
        }
        public ASN1_Gauge32(uint initValue)
            : base()
        {
            this.Gauge.Value = initValue;
        }

        [VSSerializableBinary(100)]
        public BER_UInteger Gauge = new BER_UInteger();

        public override string ToString()
        {
            return this.Gauge.ToString();
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.TimeTicks)]
    public class ASN1_Timeticks : ASN1_ValueBase
    {
        public ASN1_Timeticks()
            : base()
        {
        }
        public ASN1_Timeticks(uint initValue)
            : base()
        {
            this.TimeticksRaw = initValue;
        }
        public ASN1_Timeticks(TimeSpan initValue)
            : base()
        {
            this.Timeticks = initValue;
        }

        [VSSerializableBinary(100)]
        private BER_UInteger timeticks = new BER_UInteger();

        public TimeSpan Timeticks
        {
            get { return new TimeSpan((long)this.timeticks.Value * 100000); }
            set { this.timeticks.Value = (uint)value.Ticks / 100000; }
        }

        public uint TimeticksRaw
        {
            get { return this.timeticks.Value; }
            set { this.timeticks.Value = value; }
        }

        public override string ToString()
        {
            return string.Format("{0} ({1})", this.Timeticks.ToString(), this.TimeticksRaw);
        }
    }

    [VSSerializableContainer]
    public abstract class ASN1_ErrorBase : ASN1_EmptyBase
    {
        public override bool IsErrorIndicator
        {
            get { return true; }
        }
        public override string ToString()
        {
            return "Error Indicator";
        }
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.NoSuchObject)]
    public class ASN1_NoSuchObject : ASN1_ErrorBase
    {
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.NoSuchInstance)]
    public class ASN1_NoSuchInstance : ASN1_ErrorBase
    {
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.EndOfMIBView)]
    public class ASN1_EndOfMIBView : ASN1_ErrorBase
    {
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Sequence)]
    public class SNMP_Varbind
    {
        public SNMP_Varbind()
            : base()
        {
        }

        public SNMP_Varbind(string oid)
            : base()
        {
            this.ObjId = new ASN1_ObjId(oid);
            this.ValueObject = new ASN1_Null();
        }

        public SNMP_Varbind(string oid, ASN1_Base valueObject)
            : base()
        {
            this.ObjId = new ASN1_ObjId(oid);
            this.ValueObject = valueObject;
        }

        [VSSerializableBinary(100, CounterMode = VSCounterMode.BytesStored, CounterMemberSpan = 2)]
        private BER_Length length = new BER_Length();

        [VSSerializableSubContainer(200)]
        public ASN1_ObjId ObjId;

        [VSSerializableSubContainer(300)]
        [VSDeserializeCandidate(typeof(ASN1_Boolean), ApplyOrder = 1)]
        [VSDeserializeCandidate(typeof(ASN1_Integer), ApplyOrder = 2)]
        [VSDeserializeCandidate(typeof(ASN1_OctetString), ApplyOrder = 3)]
        [VSDeserializeCandidate(typeof(ASN1_Null), ApplyOrder = 4)]
        [VSDeserializeCandidate(typeof(ASN1_ObjId), ApplyOrder = 5)]
        [VSDeserializeCandidate(typeof(ASN1_IpAddress), ApplyOrder = 6)]
        [VSDeserializeCandidate(typeof(ASN1_Counter32), ApplyOrder = 7)]
        [VSDeserializeCandidate(typeof(ASN1_Gauge32), ApplyOrder = 8)]
        [VSDeserializeCandidate(typeof(ASN1_Timeticks), ApplyOrder = 9)]
        [VSDeserializeCandidate(typeof(ASN1_NoSuchObject), ApplyOrder = 11)]
        [VSDeserializeCandidate(typeof(ASN1_NoSuchInstance), ApplyOrder = 12)]
        [VSDeserializeCandidate(typeof(ASN1_EndOfMIBView), ApplyOrder = 13)]
        public ASN1_Base ValueObject;

        public override string ToString()
        {
            return string.Format("{0} ({1}) {2}",
                this.ObjId.OID.ToString(),
                this.ValueObject.GetType().Name,
                this.ValueObject.ToString());
        }
    }

    public enum SNMP_PDUType : byte
    {
        Get = 0xA0,
        GetNext = 0xA1,
        Response = 0xA2,
        Set = 0xA3,
        [Obsolete("Obsolete in v2, use TrapSNMPv2 instead")]
        Trap = 0xA4, //obsolete in SNMPv2
        GetBulkRequest = 0xA5,
        InformRequest = 0xA6,
        TrapSNMPv2 = 0xA7
    }

    [VSSerializableContainer(IdHeader = (byte)BER_Tag.Sequence)]
    public class SNMPv2C_Datagram
    {
        public SNMPv2C_Datagram()
            : base()
        {
        }

        public SNMPv2C_Datagram(SNMP_PDUType pduType, int id)
            : base()
        {
            this.PDUType = pduType;
            this.Id.IntVal.Value = id;
        }

        [VSSerializableBinary(100, CounterMode = VSCounterMode.BytesStored, CounterMemberSpan = 10)]
        private BER_Length length_outer = new BER_Length();

        [VSSerializableSubContainer(200)]
        public ASN1_Integer Version = new ASN1_Integer(1);

        [VSSerializableSubContainer(300)]
        public ASN1_OctetString Community = new ASN1_OctetString("public");

        [VSSerializableBinary(400)]
        public SNMP_PDUType PDUType;

        [VSSerializableBinary(500, CounterMode = VSCounterMode.BytesStored, CounterMemberSpan = 6)]
        private BER_Length length_inner = new BER_Length();

        [VSSerializableSubContainer(600)]
        private ASN1_Integer id = new ASN1_Integer(0);

        [VSSerializableSubContainer(700)]
        private ASN1_Integer int_1 = new ASN1_Integer(0);

        [VSSerializableSubContainer(800)]
        public ASN1_Integer int_2 = new ASN1_Integer(0);

        [VSSerializableBinary(900)] //TODO: Problem bei "DefaultIdentifies = true" !!??
        private BER_Tag tag_Varbinds = BER_Tag.Sequence;

        [VSSerializableBinary(1000, CounterMode = VSCounterMode.BytesStored)]
        private BER_Length length_varbinds = new BER_Length();

        [VSSerializableBinary(1100)]
        public List<SNMP_Varbind> Varbinds = new List<SNMP_Varbind>();

        public ASN1_Integer Id
        {
            get { return this.id; }
        }
        public SNMP_ErrorStatus ErrorStatus
        {
            get { return (SNMP_ErrorStatus)this.int_1.IntVal.Value; }
        }
        public ASN1_Integer ErrorIndex
        {
            get { return this.int_2; }
        }
        public ASN1_Integer NonRepeaters
        {
            get { return this.int_1; }
        }
        public ASN1_Integer MaxRepetitions
        {
            get { return this.int_2; }
        }
        public override string ToString()
        {
            return string.Format("PDUType {0} Id {1} Error({2}/{3}) Varbinds {4}",
                this.PDUType.ToString(), this.Id.IntVal, this.ErrorStatus, this.ErrorIndex.IntVal, Varbinds.Count);
        }

        public string Print()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine(this.ToString());
            foreach (SNMP_Varbind vb in this.Varbinds)
                sb.AppendLine(vb.ToString());
            return sb.ToString();
        }
    }

    /// <summary>
    /// Represents a method that is called when a snmp datagram (either a request or a trap) has been
    /// received.
    /// </summary>
    /// <remarks><note>Note that this delegate executes in a separate thread context.</note></remarks>
    /// <param name="sender">The SNMPc2C object that received the datagram.</param>
    /// <param name="datagram">The SNMPv2C_Datagram representing the request or trap.</param>
    /// <param name="source">The IPEndPoint identifying the sender.</param>
    /// <param name="reply">Optionally receives a answer datagram to be sent back to the requester.</param>
    public delegate void SNMPv2C_ReceivedDelegate(SNMPv2C sender, SNMPv2C_Datagram datagram, IPEndPoint source, out SNMPv2C_Datagram reply);

    /// <summary>
    /// Represents a method that is called when an exception occurs in an internal threaded callback method.
    /// </summary>
    /// <remarks>
    /// <para>The method must handle the passed exception in an appropriate way. After returning
    /// from this delegate method, execution is being continued as possible, unless the delegate
    /// method decides to re-throw the exception.</para>
    /// <note>Note that this delegate executes in a separate thread context.</note></remarks>
    /// <param name="sender">The SNMPc2C object where the exception occurred.</param>
    /// <param name="exception">The Exception object.</param>
    public delegate void SNMPv2C_HandleException(SNMPv2C sender, Exception exception);

    /// <summary>
    /// Represents a SNMP V2C manager or agent.
    /// </summary>
    /// <example>
    /// /* Example demonstrating a simple SNMP Get request */
    /// SNMPv2C snmp = new SNMPv2C();
    /// SNMPv2C_Datagram snmp_req = new SNMPv2C_Datagram(SNMP_PDUType.Get, snmp.GetNextReqId());
    /// snmp_req.Varbinds.Add(new SNMP_Varbind("1.3.6.1.2.1.1.5.0")); //sysName
    /// SNMPv2C_Datagram snmp_resp = snmp.Request(snmp_req, "localhost", 161, 10000);
    /// if (snmp_resp != null) Console.Out.WriteLine(snmp_resp.Print());
    /// snmp.Dispose();
    /// </example>
    public class SNMPv2C : IDisposable
    {
        #region Constructors
        /// <summary>
        /// Initializes a new instance of the SNMPv2C class without the ability to
        /// receive requests or traps.
        /// </summary>
        public SNMPv2C()
            : base()
        {
            this.factory = new VSFactory();
            this.udpClientOutgoing = new UdpClient();
        }
        /// <summary>
        /// Initializes a new instance of the SNMPv2C class and specify port numbers
        /// for receiving requests and traps.
        /// </summary>
        /// <param name="localAgentPort">Specifies the UDP port number where the object
        /// listens for incoming snmp requests. Usually this would be port number 161
        /// but it is possible to specify a differnt port number. A value of zero or below
        /// diables the request receiver.</param>
        /// <param name="localTrapPort">Specifies the UDP port number where the object
        /// listens for incoming snmp traps. Usually this would be port number 162
        /// but it is possible to specify a differnt port number. A value of zero or below
        /// diables the trap receiver.</param>
        /// <remarks><para>To initialize an instance of this class used as a snmp manager (but not
        /// as an agent) with the ability to receive traps at the standard port, one would
        /// specify 0 for localAgentPort and 162 for localTrapPort.</para>
        /// <para>Specify the localAgentPort as >0 only if you intend to implement an agent.</para>
        /// </remarks>
        /// <exception cref="SocketException">Occurs when a udp client cannot be created on the
        /// specified port.</exception>
        public SNMPv2C(int localAgentPort, int localTrapPort)
            : base()
        {
            this.factory = new VSFactory();
            this.udpClientOutgoing = new UdpClient();
            if (localAgentPort > 0)
            {
                this.localAgentPort = localAgentPort;
                this.udpClientRequests = new UdpClient(localAgentPort);
                this.udpClientRequests.BeginReceive(dataReceivedCallback, this.udpClientRequests);
            }
            if (localTrapPort > 0)
            {
                this.localTrapPort = localTrapPort;
                this.udpClientTraps = new UdpClient(localTrapPort);
                this.udpClientTraps.BeginReceive(dataReceivedCallback, this.udpClientTraps);
            }
        }
        #endregion

        #region Private Fields
        private string community = "public";
        private volatile bool disposed;
        private VSFactory factory;
        private volatile bool ignoreSocketExceptions = true;
        private object lockLastReqId = new object();
        private int lastReqId;
        private int localAgentPort;
        private int localTrapPort;
        private string remoteHostname;
        private int remotePort;
        private int timeoutMilliseconds = 5000;
        private UdpClient udpClientOutgoing;
        private UdpClient udpClientRequests;
        private UdpClient udpClientTraps;
        #endregion

        #region Private Methods
        private void dataReceivedCallback(IAsyncResult ar)
        {
            if (this.disposed) return;

            try
            {
                UdpClient udp = (UdpClient)ar.AsyncState;
                byte[] data = null;
                IPEndPoint remoteEP = null;
                try
                {
                    data = udp.EndReceive(ar, ref remoteEP);
                }
                catch (SocketException e)
                {
                    if (e.ErrorCode == 10054)
                    {
                        //Connection reset by peer (WSAECONNRESET)
                        //A connection was forcibly closed by the peer. This normally results from a
                        //loss of the connection on the remote socket due to a timeout or a reboot. 
                        //the underlying socket is closed now so just recreate and reconnect the udp object
                        udp.Close();
                        if (udp == this.udpClientRequests)
                        {
                            this.udpClientRequests = new UdpClient(this.localAgentPort);
                            udp = this.udpClientRequests;
                        }
                        else
                            if (udp == this.udpClientTraps)
                            {
                                this.udpClientTraps = new UdpClient(this.localTrapPort);
                                udp = this.udpClientTraps;
                            }
                            else
                                udp = null; //deet
                        data = null;
                    }
                    else
                        throw;
                }
                if ((data != null) && (data.Length > 0) && (udp != null))
                    dataReceivedEval(data, udp, remoteEP);
                if (udp != null)
                    udp.BeginReceive(dataReceivedCallback, udp);
            }
            catch (Exception e)
            {
                if (this.OnHandleException != null)
                    this.OnHandleException(this, e);
                else
                    if (!(e is SocketException) || !this.ignoreSocketExceptions)
                        throw;
            }
        }
        private void dataReceivedEval(byte[] data, UdpClient fromUdp, IPEndPoint remoteEP)
        {
            SNMPv2C_Datagram dg;
            lock (this.factory)
                dg = this.factory.DeserializeOne<SNMPv2C_Datagram>(data, false);
            if ((dg != null) && (this.OnReceived != null))
            {
                SNMPv2C_Datagram dg_reply;
                this.OnReceived(this, dg, remoteEP, out dg_reply);
                if (dg_reply != null)
                    this.sendDatagram(dg_reply, fromUdp, remoteEP);
            }
        }
        private bool sendDatagram(SNMPv2C_Datagram datagram, UdpClient udpClient, IPEndPoint destEP)
        {
            byte[] data;
            lock (this.factory)
                data = this.factory.SerializeToBytes(datagram);
            if ((data == null) || (data.Length == 0))
                return false;
            lock (udpClient)
                udpClient.Send(data, data.Length, destEP);
            return true;
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Gets a new request id to assign to <see cref="SNMPv2C_Datagram.Id" />.
        /// </summary>
        /// <returns>This method is implemented thread-safe.</returns>
        public int GetNextReqId()
        {
            lock (this.lockLastReqId)
            {
                if (this.lastReqId == int.MaxValue) this.lastReqId = 0;
                return ++this.lastReqId;
            }
        }
        /// <summary>
        /// Sends a snmp request to a remote computer identified by a hostname and a port nummber.
        /// </summary>
        /// <param name="request">The SNMPv2C_Datagram that represents the request to send.</param>
        /// <param name="hostname">Name or ip address of the addressed computer.</param>
        /// <param name="port">Port number to address.</param>
        /// <param name="timeoutMilliseconds">A timeout value in milliseconds specifying the
        /// maximal time to wait for a response on the request.</param>
        /// <returns>A SNMPv2C_Datagram object representing the reponse or null if no valid reponse
        /// was received.</returns>
        public SNMPv2C_Datagram Request(SNMPv2C_Datagram request, string hostname, int port, int timeoutMilliseconds)
        {
            SNMPv2C_Datagram retval = null;
            try
            {
                if (this.SendDatagram(request, hostname, port))
                {
                    IAsyncResult ar;
                    lock (this.udpClientOutgoing)
                        ar = this.udpClientOutgoing.BeginReceive(null, null);

                    while (true)
                    {
                        ar.AsyncWaitHandle.WaitOne(timeoutMilliseconds, false);
                        byte[] resp_data;
                        IPEndPoint ep = null;
                        lock (this.udpClientOutgoing)
                            resp_data = this.udpClientOutgoing.EndReceive(ar, ref ep);
                        if (resp_data != null)
                            lock (this.factory)
                                retval = this.factory.DeserializeOne<SNMPv2C_Datagram>(resp_data, false);
                        else
                            retval = null;
                        if ((retval == null) || (retval.Id.IntVal.Value == request.Id.IntVal.Value))
                            break;
                    }
                }
            }
            catch (SocketException)
            {
                if (this.ignoreSocketExceptions)
                    retval = null;
                else
                    throw;
            }

            return retval;
        }
        /// <summary>
        /// Sends a snmp request.
        /// </summary>
        /// <param name="request">The SNMPv2C_Datagram that represents the request to send.</param>
        /// <remarks>The remote address and port and the timeout value are specified by the
        /// properties <see cref="RemoteHostname" />, <see cref="RemotePort" /> and
        /// <see cref="TimeoutMilliseconds" />.</remarks>
        /// <returns>A SNMPv2C_Datagram object representing the reponse or null if no valid reponse
        /// was received.</returns>
        public SNMPv2C_Datagram Request(SNMPv2C_Datagram request)
        {
            return this.Request(request, this.remoteHostname, this.remotePort, this.timeoutMilliseconds);
        }
        /// <summary>
        /// Sends a SNMPv2C_Datagram to the partner addessed using an IPEndPoint,
        /// without waiting for a response.
        /// </summary>
        /// <param name="datagram">The SNMPv2C_Datagram to send.</param>
        /// <param name="destEP">The IPEndPoint identifying the receiver.</param>
        /// <returns>A bool value indicating whether the sending the datagram was successful. Note that
        /// a return value of true does not guarantee that the partner receives and understands the sending.</returns>
        public bool SendDatagram(SNMPv2C_Datagram datagram, IPEndPoint destEP)
        {
            return this.sendDatagram(datagram, this.udpClientOutgoing, destEP);
        }
        /// <summary>
        /// Sends a SNMPv2C_Datagram to the partner addessed by its host name and port number,
        /// without waiting for a response.
        /// </summary>
        /// <param name="datagram">The SNMPv2C_Datagram to send.</param>
        /// <param name="hostname">The host name or ip address of the partner.</param>
        /// <param name="port">The port number of the partner.</param>
        /// <returns>A bool value indicating whether the sending the datagram was successful. Note that
        /// a return value of true does not guarantee that the partner receives and understands the sending.</returns>
        public bool SendDatagram(SNMPv2C_Datagram datagram, string hostname, int port)
        {
            IPAddress[] ips = Dns.GetHostAddresses(hostname);
            foreach (IPAddress ip in ips)
            {
                if (ip.AddressFamily == AddressFamily.InterNetwork)
                    return this.SendDatagram(datagram, new IPEndPoint(ip, port));
            }
            return false;
        }

        /// <summary>
        /// Performs a "walk", i.e. a series of consecutive get-bulk requests, over the mib of a remote
        /// computer identified by a hostname and a port nummber.
        /// </summary>
        /// <param name="startOid">The starting point of the walk.</param>
        /// <param name="maxRepetitions">The number of varbinds to request for a single bulk-response.
        /// The value must be greater than zero.</param>
        /// <param name="hostname">Name or ip address of the addressed computer.</param>
        /// <param name="port">Port number to address.</param>
        /// <param name="timeoutMilliseconds">A timeout value in milliseconds specifying the
        /// maximal time to wait for a response on a single request.</param>
        /// <param name="errorStatus">Receives the error status reported in the last reponse.</param>
        /// <remarks>
        /// The first response with an error status different from <see cref="SNMP_ErrorStatus.noError" />
        /// aborts the walk and the errorStatus parameter receives this value.</remarks>
        /// <returns>A list of all SNMP_Varbind objects received during the walk or null if an error
        /// occurred.</returns>
        public List<SNMP_Varbind> Walk(string startOid, int maxRepetitions, string hostname, int port, int timeoutMilliseconds, out SNMP_ErrorStatus errorStatus)
        {
            errorStatus = SNMP_ErrorStatus.noError;
            List<SNMP_Varbind> retval = new List<SNMP_Varbind>();

            string continue_oid = startOid;

            while (continue_oid != null)
            {
                SNMPv2C_Datagram snmp_req = new SNMPv2C_Datagram(SNMP_PDUType.GetBulkRequest, this.GetNextReqId());
                snmp_req.MaxRepetitions.IntVal.Value = maxRepetitions;
                snmp_req.Varbinds.Add(new SNMP_Varbind(continue_oid));

                SNMPv2C_Datagram snmp_resp = this.Request(snmp_req, hostname, port, timeoutMilliseconds);
                if (snmp_resp == null)
                    return null;
                errorStatus = snmp_resp.ErrorStatus;
                continue_oid = null;
                if (errorStatus == SNMP_ErrorStatus.noError)
                {
                    for (int i = 0; i < snmp_resp.Varbinds.Count; i++)
                    {
                        if (!snmp_resp.Varbinds[i].ValueObject.IsErrorIndicator &&
                            (continue_oid = snmp_resp.Varbinds[i].ObjId.OID.Value).StartsWith(startOid))
                        {
                            retval.Add(snmp_resp.Varbinds[i]);
                        }
                        else
                        {
                            continue_oid = null;
                            break;
                        }
                    }
                }
            }

            return retval;
        }
        /// <summary>
        /// Performs a "walk", i.e. a series of consecutive get-bulk requests.
        /// </summary>
        /// <param name="startOid">The starting point of the walk.</param>
        /// <param name="maxRepetitions">The number of varbinds to request for a single bulk-response.
        /// The value must be greater than zero.</param>
        /// <param name="errorStatus">Receives the error status reported in the last reponse.</param>
        /// <remarks>
        /// <para>The first response with an error status different from <see cref="SNMP_ErrorStatus.noError" />
        /// aborts the walk and the errorStatus parameter receives this value.</para>
        /// <para>The remote address and port and the timeout value are specified by the
        /// properties <see cref="RemoteHostname" />, <see cref="RemotePort" /> and
        /// <see cref="TimeoutMilliseconds" />.</para></remarks>
        /// <returns>A list of all SNMP_Varbind objects received during the walk or null if an error
        /// occurred.</returns>
        public List<SNMP_Varbind> Walk(string startOid, int maxRepetitions, out SNMP_ErrorStatus errorStatus)
        {
            return this.Walk(startOid, maxRepetitions, this.remoteHostname, this.remotePort, this.timeoutMilliseconds, out errorStatus);
        }
        #endregion

        #region Public Properties
        /// <summary>
        /// Gets or sets the community name. If not specified, it defaults to "public".
        /// </summary>
        public string Community
        {
            get { return this.community; }
            set { this.community = value; }
        }
        /// <summary>
        /// Specifies whether or not the occurrence of any SocketException shall be ignored.
        /// The default value is true;
        /// </summary>
        /// <seealso cref="OnHandleException"/>
        public bool IgnoreSocketExceptions
        {
            get { return this.ignoreSocketExceptions; }
            set { this.ignoreSocketExceptions = value; }
        }
        /// <summary>
        /// Fires when either an exception occurred in an internal threaded callback method during receiving
        /// unsolicited requests or traps.
        /// </summary>
        /// <remarks>
        /// <para>This event can only occur when either localAgentPort or localTrapPort
        /// was specified > 0 using this <see cref="SNMPv2C(int,int)" /> constructor of this class.</para>
        /// <para>Even if <see cref="IgnoreSocketExceptions" /> is set to true, exceptions of the class
        /// SocketException are propagated using this event. However, if this event is not assigned and
        /// an exception other than SocketException occurs in an internal callback or even a
        /// SocketException occurs but <see cref="IgnoreSocketExceptions" /> to set to false,
        /// execution of the process might cease due to the exception being unhandled.</para>
        /// <note>The event handler might be executed in the context of a separate thread.</note>
        /// </remarks>
        public event SNMPv2C_HandleException OnHandleException;
        /// <summary>
        /// Occurs when either a SNMP request or a trap was received.
        /// </summary>
        /// <remarks>
        /// <para>This event can only occur when either localAgentPort or localTrapPort
        /// was specified > 0 using this <see cref="SNMPv2C(int,int)" /> constructor of this class.</para>
        /// <note>The event handler might be executed in the context of a separate thread.</note>
        /// </remarks>
        public event SNMPv2C_ReceivedDelegate OnReceived;
        /// <summary>
        /// Gets or sets the default hostname of the partner host.
        /// </summary>
        /// <remarks>
        /// This address is used when the methods Request(SNMPv2C_Datagram) or
        /// Walk(string,int,out SNMP_ErrorStatus) are called.
        /// </remarks>
        /// <seealso cref="RemotePort"/>
        public string RemoteHostname
        {
            get { return this.remoteHostname; }
            set { this.remoteHostname = value; }
        }
        /// <summary>
        /// Gets or sets the default port number of the partner host.
        /// </summary>
        /// <remarks>
        /// This address is used when the methods Request(SNMPv2C_Datagram) or
        /// Walk(string,int,out SNMP_ErrorStatus) are called.
        /// </remarks>
        /// <seealso cref="RemoteHostname"/>
        public int RemotePort
        {
            get { return this.remotePort; }
            set { this.remotePort = value; }
        }
        /// <summary>
        /// Gets or sets the default timeout value in milliseconds to wait for a response on a request.
        /// The default value is 5000.
        /// </summary>
        /// <remarks>
        /// <para>This value is used when the methods Request(SNMPv2C_Datagram) or Walk(string,int) are called.</para>
        /// <para>Specifying a value of Timeout.Infinite (-1) might cause a request or walk to block forever.</para>
        /// </remarks>
        public int TimeoutMilliseconds
        {
            get { return this.timeoutMilliseconds; }
            set { this.timeoutMilliseconds = value; }
        }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Releases resources used by the object. 
        /// </summary>
        public void Dispose()
        {
            if (!disposed)
            {
                disposed = true;
                if (this.udpClientOutgoing != null)
                {
                    ((IDisposable)this.udpClientOutgoing).Dispose();
                    this.udpClientOutgoing = null;
                }
                if (this.udpClientRequests != null)
                {
                    ((IDisposable)this.udpClientRequests).Dispose();
                    this.udpClientRequests = null;
                }
                if (this.udpClientTraps != null)
                {
                    ((IDisposable)this.udpClientTraps).Dispose();
                    this.udpClientTraps = null;
                }
            }
        }
        #endregion
    }

#pragma warning restore 414 //warning field is assigned but its value is never used
}
